/**
 * Copyright (C) 2022 Xilinx, Inc
 * Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

// ------ I N C L U D E   F I L E S -------------------------------------------
// Local - Include Files
#include "ReportDeviceHW.h"
#include "MAUtilities.h"

// 3rd Party Library - Include Files
#include <string>
#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/property_tree/json_parser.hpp>

#define BYTES_TO_MEGABYTES 0x100000ll

void ReportDeviceHW::getPropertyTree(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& _pt) const {
  boost::property_tree::ptree pt;

  XBUtilities::get_device_info(dev, pt);
  // There can only be 1 root node
  _pt.add_child("device_info", pt);
}

void ReportDeviceHW::writeReport(const boost::property_tree::ptree& _pt, std::ostream& _output) const {
  boost::property_tree::ptree empty_ptree;

  _output << "Device Hardware Info:\n";
  if (_pt.get<std::string>("device_info.status") != "ok") {
    _output << "ERROR " << _pt.get<std::string>("device_info.status") << std::endl;
    return;
  }

  try {
    //boost::property_tree::json_parser::write_json(std::cout, _pt, true /*Pretty Print*/);
    const auto& fw_version      = _pt.get_child("device_info.device_info.version_info", empty_ptree);
    const auto& chip_thresholds = _pt.get_child("device_info.device_info.thresholds", empty_ptree);
    const auto& chip_info       = _pt.get_child("device_info.device_info.chip_id", empty_ptree);
    const auto  uptime          = _pt.get<std::string>("device_info.device_info.uptime_S", "?");

    _output << "Device uptime (sec):" << uptime << "\n";
    _output << "Device Firmware Info:\n";
    for (const auto& kd : fw_version) {
      _output << "  " << kd.first << ":  " << kd.second.get_value<std::string>() << "\n";
    }
    _output << "Device Threshold Info:\n";
    for (const auto& kd : chip_thresholds) {
      _output << "  " << kd.first << ":  " << kd.second.get_value<std::string>() << "\n";
    }
    _output << "Device Hardware Info:\n";
    for (const auto& kd : chip_info) {
      _output << "  " << kd.first << ":  " << kd.second.get_value<std::string>() << "\n";
    }
    _output << "  Product_Name:  " << _pt.get<std::string>("device_info.device_info.product_name", "?") << "\n";
    _output << "  Product_Revision:  " << _pt.get<std::string>("device_info.device_info.product_revision", "?") << "\n";
    _output << "  Product_SN:  " << _pt.get<std::string>("device_info.device_info.product_serial_number", "?") << "\n";
    _output << "  Processor_Type:  " << _pt.get<std::string>("device_info.device_info.processor_type", "?") << "\n";
  } catch (const boost::property_tree::ptree_error& ex) {
    _output << "ERROR: Unexpected error" << std::endl;
  }
  _output << std::endl;
}
