// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <jsf_helpers.h>

class Device {
private:
  // Telemetry
  static constexpr uint32_t MAILBOX1_OFFSET = 0x800;
  static constexpr uint32_t S1_SYS_CON_OFF  = 0x0190'0000;
  static constexpr uint32_t MAILBOX2_OFFSET = 0x860;
  static constexpr uint32_t S2_SYS_CON_OFF  = 0x0280'0000;

  // static constexpr uint32_t PVT_P_NUM  = 6;
  // static constexpr uint32_t PVT_NUM    = 4;
  static constexpr uint32_t LPDDR5_DATA_RATE = 5500;
  static constexpr uint32_t DDR_CTRL_FREQ    = LPDDR5_DATA_RATE / 8; // for LPDDR5

private:
  std::string m_dbdf;
  pci_dev     m_pciDevice;
  bool        m_demo;

public:
  static constexpr uint32_t SRAM_SC_OFF     = 0x0290'0000;
  static constexpr uint32_t ZSP_SRAM_ADDR   = SRAM_SC_OFF;
  static constexpr uint32_t CMD_DIR_ADDR    = 0x1fff4;
  static constexpr uint32_t CMD_RSN_ADDR    = 0x1fff8;
  static constexpr uint32_t SIGNATURE_TO_RC = 0x544F'5243;
  static constexpr uint32_t RUN_SUCCESS     = 0x8385'6767;

  // Mailbox Struct Offsets
  static constexpr uint32_t MAILBOX_RESERVE_SIZE = 0x20000;
  static constexpr uint32_t MAILBOX_BASE_OFFSET  = 0x80000;

  static constexpr uint32_t VERSION_INFO   = MAILBOX_BASE_OFFSET;
  static constexpr uint32_t EVENT_ID       = VERSION_INFO + 0x18;
  static constexpr uint32_t PROCESS_STATUS = EVENT_ID + 0x4;
  static constexpr uint32_t DDR_BW_INFO    = PROCESS_STATUS + 0x4;
  static constexpr uint32_t UPTIME_S       = DDR_BW_INFO + 0xa0;

  static constexpr uint32_t PVT_INFO        = UPTIME_S + 0x8;
  static constexpr uint32_t PVT_VOLTAGE     = PVT_INFO + 0x48;
  static constexpr uint32_t PVT_TEMPERATURE = PVT_VOLTAGE + 0x10;

  static constexpr uint32_t ERROR_INFO                 = PVT_INFO + 0x68;
  static constexpr uint32_t PCIE_ERROR_COUNTER         = PVT_INFO + 0x68;
  static constexpr uint32_t DDR_ECC_ERROR_COUNTER      = PCIE_ERROR_COUNTER + 0x4;
  static constexpr uint32_t AXI_SRAM_ECC_ERROR_COUNTER = DDR_ECC_ERROR_COUNTER + 0x20;

  static constexpr uint32_t BOARD_TEMP = ERROR_INFO + 0x2c;
  static constexpr uint32_t BOARD_VC   = BOARD_TEMP + 0x4;

  static constexpr uint32_t EEPROM_DATA_DEVICE_INFO = BOARD_VC + 0xc;
  static constexpr uint32_t PRODUCT_INFO            = BOARD_VC + 0xc;
  static constexpr uint32_t NAME                    = BOARD_VC + 0xc;
  static constexpr uint32_t REVISION                = NAME + 0x18;
  static constexpr uint32_t SERIAL_NUMBER           = REVISION + 0x8;
  static constexpr uint32_t ID_INFO                 = EEPROM_DATA_DEVICE_INFO + 0x30;
  static constexpr uint32_t PART_NUMBER             = EEPROM_DATA_DEVICE_INFO + 0x30;
  static constexpr uint32_t OEM_ID                  = PART_NUMBER + 0x18;
  static constexpr uint32_t PCIE_INFO               = OEM_ID + 0x4;
  static constexpr uint32_t MAX_POWER               = PCIE_INFO + 0x8;

  static constexpr uint32_t ASIC_NUM = EEPROM_DATA_DEVICE_INFO + 0x58;

  static constexpr uint32_t FLASH_STATUS     = ASIC_NUM + 0x4;
  static constexpr uint32_t SC_UPDATE_STATUS = FLASH_STATUS + 0x4;

  static constexpr uint32_t VRM_VC      = SC_UPDATE_STATUS + 0x4;
  static constexpr uint32_t ADC_VOLTAGE = VRM_VC + 0x8;
  static constexpr uint32_t VRM_TEMP    = ADC_VOLTAGE + 0x10;

  static constexpr uint32_t SC_UPDATE_FLAG = MAILBOX_BASE_OFFSET + MAILBOX_RESERVE_SIZE + 0x30;

  int         m_device;
  uint64_t    m_bar2Base;
  uint64_t    m_bar2Size;
  void*       m_bar2Vaddr;
  uint64_t    m_bar4Base;
  uint64_t    m_bar4Size;
  void*       m_bar4Vaddr;
  std::string m_iommu_grp;
  unsigned    m_deviceNum;

public:
  explicit Device(const pci_dev& PciDevice, bool Demo);
  virtual ~Device() = default;

  template <typename T = uint32_t>
  T ReadBar2(uint64_t Offset) const;

  template <typename T = uint32_t>
  unsigned WriteBar2(uint64_t Offset, const T& Val);

  template <typename T = uint32_t>
  T ReadBar4(uint64_t Offset) const;

  template <typename T = uint32_t>
  unsigned WriteBar4(uint64_t Offset, const T& Val);

  pci_dev     GetPciDevice() { return m_pciDevice; }
  std::string GetBdf() { return m_dbdf; }
  bool        IsDemo() { return m_demo; }
  unsigned    GetDeviceNum() { return m_deviceNum; }

  void fail() { exit(1); }

  uint32_t    GetDeviceVoltage(unsigned Index, unsigned Offset) const;
  uint32_t    GetDeviceTemperature(unsigned Index, unsigned Offset) const;
  uint32_t    GetSensorInfo(unsigned Index, unsigned Offset) const;
  uint32_t    GetBoardVcInfo(unsigned Index, unsigned Offset) const;
  uint32_t    GetUptime(unsigned Offset) const;
  std::string GetVersion(unsigned Index, unsigned Offset) const;
  std::string GetProductInfo(unsigned Index, unsigned Offset) const;
  std::string GetPcieInfo(const std::string& Val) const;

  unsigned long                         GetDdrInfo(unsigned Index, unsigned Offset) const;
  std::pair<unsigned int, unsigned int> GetErrorCounters(unsigned Index, unsigned Offset) const;

  static std::unique_ptr<Device> CreateDevice(const pci_dev& PciDevice, bool Demo);
};
